
library fields.field_editor;

import 'dart:async';
import 'dart:mirrors';
import 'field.dart';
import 'theme.dart';
import 'transform.dart';
import 'src/mirrors.dart';

class Resources {
  String get style =>
"""
input {
  background-color:transparent;
  border-style:none;
  color:${theme.colors.text};
}
input[type="text"] {  
  width:100%;
}
input[type="text"][disabled] { 
  color:${theme.colors.textDisabled};
}
input[class=error] {
  border: 2px solid ${theme.colors.error};
}
""";
  const Resources();
}

abstract class FieldEditor {
  static final _fieldEditorMirror = reflectClass(FieldEditor);
  static final _editors = new Map<Type, ClassMirror>();
  static ClassMirror _default;
    
  static bool has(Type fieldType) => _editors.containsKey(fieldType);
  static bool get hasDefault => _default != null;
  
  static void register(Type fieldType, Type editorType) {
    if (editorType == null) 
      throw new ArgumentError('editorType must be !null');
    if (has(fieldType)) 
      throw new StateError('$fieldType already has a registered editorType');    
    final editorClassMirror = reflectClass(editorType);
    if (!isAssignableFrom(_fieldEditorMirror, editorClassMirror))
      throw new ArgumentError(
          'editorType must extend or implement FieldEditor');
    _editors[fieldType] = editorClassMirror;
  }
  
  static void registerDefault(Type editorType) {
    if (editorType == null) 
      throw new ArgumentError('editorType must be !null');
    if (hasDefault) 
      throw new StateError('FieldEditor has default editorType.');
    final editorClassMirror = reflectClass(editorType);
    if (!isAssignableFrom(_fieldEditorMirror, editorClassMirror))
      throw new ArgumentError(
          'editorType must extend or implement FieldEditor');
    _default = editorClassMirror;
  }
  
  static bool unregister(Type fieldType) => _editors.remove(fieldType) != null;
  
  static bool unregisterDefault() {
    final has = hasDefault;
    _default = null;
    return has;
  }
  
  static void unregisterAll() {
    unregisterDefault();
    _editors.clear();
  }
    
  /*Element*/ get element;
  
  final Field field;
  final Resources resources;
  
  Type get valueType;
  
  bool get isValueReadOnly => 
      field.isReadOnly || !canTransform(valueType, field.type);
  
  factory FieldEditor.forField(Field field) {
    ClassMirror mirror = _default;
    if (field.editor != null)
      mirror = field.editor;
    else if (has(field.type)) 
      mirror = _editors[field.type];
    if (mirror == null) 
      throw new StateError('No editor registered for type <${field.type}>');
    return mirror.newInstance(const Symbol(''),[field]).reflectee;
  }
  
  FieldEditor(this.field, [this.resources = const Resources()]);
    
  Future getValue() { 
    return field.getValue().then((value) {
      if(canTransform(field.type, valueType))
        return new Future.value(transform(value, field.type, valueType));
      else return new Future.value(value.toString());
    });
  }

  Future setValue(value) {
    if(isValueReadOnly) {
      return new Future.error(
          new UnsupportedError('Cannot set a read-only value'));
    }
    var transformed;
    try {
      transformed = transform(value, valueType, field.type);      
    } on TransformError catch(e) {
      return new Future.error(e);
    }
    return field.setValue(transformed);
  }
}
