
library fields.object_editor;

import 'dart:async';
import 'package:tags/builder.dart';
import 'package:tags/custom.dart';
import 'package:tags/html.dart';
import 'field.dart';
import 'field_list.dart';
import 'theme.dart';
import 'src/default_editors.dart';
import 'src/field_list_view.dart';

final Tag objectEditor = 
    ([Build b, Resources res]) => 
        build(() => new ObjectEditor(res).element, b);

class Resources {
  String get style =>
"""
@host {
  * {
    background-color: ${theme.colors.background};
    border:1px solid ${theme.colors.border};
    height:100%;    
    overflow:hidden;
    width:100%;
  }
}
.toggle {
  background-color:transparent;
  border-style:none;
  float:left;
}
.toggle:disabled {
  opacity:0.4;
}
.filterbox {  
  display:block; 
  overflow:hidden;
}
.filter {
  background-color:transparent;
  border-style:none;
  color:${theme.colors.text};
  width:100%;
}
""";

  String get down => theme.images.down;
  String get up => theme.images.up;
  String get lock => theme.images.lock;
  String get unlock => theme.images.unlock;  
  const Resources();
}

class ObjectEditor implements HasElement {

  String _filterText = '';  
  Predicate _filter;
  Predicate _filterPrivate;
  
  Element _element;
  Element get element => _element;
  
  ButtonElement _sortButton;
  ButtonElement _lockButton;
  InputElement _filterBox;
  
  FieldList _list;
  FieldListView _listView;
  StreamSubscription _listListener;
  
  bool get hasSelected => _selected != null;
      
  Object _selected;
  Object get selected => _selected;
  set selected(Object value) {           
    if(value != selected) {
      if (_listListener != null) _listListener.cancel();      
      _selected = value;
      _list = new FieldList()
          ..refreshSource(_selected)
          ..predicate = _filterPrivate;      
      _listListener = _list.fieldChanged.listen((_) {
        if (_selectedChanged.hasListener) _selectedChanged.add(_selected);
      });
      _listView.clearCache();
      _render();
      if (_selectedChanged.hasListener) _selectedChanged.add(_selected);
    }
  }
  
  final _selectedChanged = new StreamController<Object>.broadcast();
  Stream<Object> get selectedChanged => _selectedChanged.stream;
  
  ObjectEditor([Resources res = const Resources()]) {
    _filter = (Field f) => f.displayName.contains(_filterText);
    _filterPrivate = (Field f) => !f.isPrivate && _filter(f);
    _element = 
        div(() {
          shadow(() {
            style()..text = res.style;
            _sortButton = toggleButton(() {
              img()..src = res.up..classes = ['up'];
              img()..src = res.down..classes = ['down'];            
            }, (ToggleButton toggle) {
              toggle.states.listen((down) {
                if (down) {
                  _list.comparator = 
                      (Field a, Field b) => a.displayName.compareTo(b.displayName) * -1;
                } else {
                  _list.comparator = FieldList.DEFAULT_COMPARATOR;
                }
                _render();
              });
            })..classes = ['toggle']
              ..disabled = true;
            _lockButton = toggleButton(() {
              img()..src = res.lock..classes = ['up'];
              img()..src = res.unlock..classes = ['down'];
            }, (ToggleButton toggle) {
              toggle.states.listen((down) {
                if (down) _list.predicate = _filter;
                else _list.predicate = _filterPrivate;
                _render();
              });
            })..classes = ['toggle']
              ..disabled = true;
            span(() {
              _filterBox = input()
                  ..type = 'text'
                  ..classes = ['filter']
                  ..disabled = true
                  ..placeholder = 'filter fields';
              })..classes = ['filterbox']                
                ..onInput.listen((_) {
                  _filterText = _filterBox.value;
                  _list.refreshView();
                  _render();
                });
            fieldListView(null, (model) => _listView = model);
          });
        });
    
    registerDefaultEditors();
  }
  
  void refresh() {
    if (hasSelected) {
      _list.refreshSource(_selected);
      _render();
    }
  }
  
  void _render() { 
    _sortButton.disabled = !hasSelected;
    _lockButton.disabled = !hasSelected;
    _filterBox.disabled = !hasSelected;
    _listView.render(_list);
  }
}
