
library fields.test.field_editor;

import 'dart:mirrors';
import 'package:unittest/unittest.dart';
import 'package:fields/field.dart';
import 'package:fields/field_editor.dart';
import 'package:fields/transform.dart';

fieldEditorTest() {
  group('[field_editor.register]', () {
    setUp(() => FieldEditor.unregisterAll());
    test('Test that no default field editor is initially registered.', 
        testHasNoDefaultFieldEditor);    
    test('Test that register default throws if the given editor type is null.', 
        testRegisterDefaultNullThrows);
    test('Test that register throws if the given editor type is null.', 
        testRegisterNullEditorTypeThrows);
    test('Test the successful register of a default editor.', 
        testRegisterDefault);
    test('Test that a second attempt to register a default editor throws.', 
        testRegisterDefaultTwiceThrows);
    test('Test that an attempt to register default a non editor type  throws.', 
        testRegisterDefaultNonEditorTypeThrows);
    test('Test the successful register of an int editor.', 
        testRegisterIntEditor);
    test('Test that a second attempt to register an int editor throws.', 
        testRegisterIntEditorTwiceThrows);
    test('Test that an attempt to register a non editor type throws.', 
        testRegisterNonEditorTypeThrows);
  });
  group('[field_editor.factory]', () {
    setUp(() => FieldEditor.unregisterAll());
    test('Test that the factory throws if no field editor is registered.', 
        testFactoryNoEditorThrows);
    test('Test that the factory returns the default editor.', 
        testFactoryReturnsDefaultEditor);
    test('Test that the factory returns registered editors with no default.', 
        testFactoryReturnsNoDefaultEditor);
  });
  group('[field_editor.value]', () {
    setUp(() => FieldEditor.unregisterAll());
    test('Test getting the value of an int field in an int editor.', 
        testIntFieldIntEditorGetValue);
    test('Test getting the value of an int field in a string editor.', 
        testIntFieldStringEditorGetValue);
    test('Test getting the value of a string field in an int editor throws.', 
        testStringFieldIntEditorGetValueThrows);
    test('Test that the value of a rectangle field in a string editor is '
        'read-only when no string -> rectangle transform is registered.', 
        testRectagleFieldStringEditorValueReadOnly);
    test('Test setting the value of an int field in an int editor.', 
        testIntFieldIntEditorSetValue);
    test('Test setting the value of an int field in a string editor.', 
        testIntFieldStringEditorSetValue);
    test('Test setting bad value of an int field in a string editor throws.', 
        testIntFieldStringEditorSetValueThrows);
    test('Test that setting the value of a rectangle field in a string '
        'editor throws because the editor is read-only.', 
         testRectangleFieldStringEditorSetValueThrows);
  });
  group('[field_editor.resources]', () {
    setUp(() => FieldEditor.unregisterAll());
    test('Test that a rectangle editor may specify a custom resources type.', 
         testRectangleEditorResources);    
  });
}

// -----------------------------------------------------------------------------
// Setup
// -----------------------------------------------------------------------------

class Rectangle {
  int x, y, w, h;
  Rectangle(this.x, this.y, this.w, this.h);
  String toString() => 'rectangle($x,$y,$w,$h)';
}

class HasFields {
  int i = 7;
  String s = 'snarf';
  Rectangle r = new Rectangle(1, 2, 3, 4);
}

class NotAnEditor { }

class IntEditor extends FieldEditor {
  final Type valueType = int;
  final element = null;
  IntEditor(Field field) : super(field);
}

class StringEditor extends FieldEditor {
  final Type valueType = String; 
  final element = null;
  StringEditor(Field field) : super(field);
}

class RectangleResources extends Resources {
  String get style => 'SOME CSS';
  const RectangleResources();
}

class RectangleEditor extends FieldEditor {
  final Type valueType = Rectangle;
  final element = null;
  RectangleEditor(Field field) : super(field, const RectangleResources());
}

// -----------------------------------------------------------------------------
// Register
// -----------------------------------------------------------------------------

void testHasNoDefaultFieldEditor() {
  expect(FieldEditor.hasDefault, isFalse);
}

void testRegisterDefaultNullThrows() {  
  expect(() {
    FieldEditor.registerDefault(null);
  }, throwsA(isArgumentError));
}

void testRegisterNullEditorTypeThrows() {
  expect(() {
    FieldEditor.register(int, null);
  }, throwsA(isArgumentError));
}

void testRegisterDefault() {
  expect(FieldEditor.hasDefault, isFalse);
  FieldEditor.registerDefault(IntEditor);
  expect(FieldEditor.hasDefault, isTrue);
}

void testRegisterDefaultTwiceThrows() {
  expect(FieldEditor.hasDefault, isFalse);
  FieldEditor.registerDefault(IntEditor);  
  expect(() {
    FieldEditor.registerDefault(IntEditor);
  }, throwsA(isStateError));
}

void testRegisterDefaultNonEditorTypeThrows() {
  expect(FieldEditor.hasDefault, isFalse);
  expect(() {
    FieldEditor.registerDefault(NotAnEditor);
  }, throwsA(isArgumentError));
}

void testRegisterIntEditor() {
  expect(FieldEditor.has(int), isFalse);
  FieldEditor.register(int, IntEditor);
  expect(FieldEditor.has(int), isTrue);
}

void testRegisterIntEditorTwiceThrows() {
  expect(FieldEditor.has(int), isFalse);
  FieldEditor.register(int, IntEditor);  
  expect(() {
    FieldEditor.register(int, IntEditor);
  }, throwsA(isStateError));
}

void testRegisterNonEditorTypeThrows() {
  expect(FieldEditor.has(int), isFalse);
  expect(() {
    FieldEditor.register(int, NotAnEditor);
  }, throwsA(isArgumentError));
}

// -----------------------------------------------------------------------------
// Factory
// -----------------------------------------------------------------------------

void testFactoryNoEditorThrows() {
  final obj = new HasFields();
  final mirror = reflect(obj);
  final field = new Field(mirror, const Symbol('i'), int);
  expect(() {
    var editor = new FieldEditor.forField(field);
  }, throwsA(isStateError));
}

void testFactoryReturnsDefaultEditor() {
  FieldEditor.registerDefault(IntEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final i = new Field(mirror, const Symbol('i'), int);
  final s = new Field(mirror, const Symbol('s'), String);
  final iEditor = new FieldEditor.forField(i);
  final sEditor = new FieldEditor.forField(s);
  expect(iEditor, new isInstanceOf<IntEditor>());
  expect(iEditor.field, equals(i));
  expect(sEditor, new isInstanceOf<IntEditor>());
  expect(sEditor.field, equals(s));
}

void testFactoryReturnsNoDefaultEditor() {
  FieldEditor.register(int, IntEditor);
  FieldEditor.register(String, StringEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final i = new Field(mirror, const Symbol('i'), int);
  final s = new Field(mirror, const Symbol('s'), String);
  final iEditor = new FieldEditor.forField(i);
  final sEditor = new FieldEditor.forField(s);
  expect(iEditor, new isInstanceOf<IntEditor>());
  expect(iEditor.field, equals(i));
  expect(sEditor, new isInstanceOf<StringEditor>());
  expect(sEditor.field, equals(s));
}

// -----------------------------------------------------------------------------
// Value
// -----------------------------------------------------------------------------

void testIntFieldIntEditorGetValue() {
  FieldEditor.register(int, IntEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final i = new Field(mirror, const Symbol('i'), int);
  IntEditor editor = new FieldEditor.forField(i);
  expect(editor.isValueReadOnly, isFalse);
  editor.getValue().then(expectAsync1((v) {
    expect(v, equals(7));
  }));
}

void testIntFieldStringEditorGetValue() {
  FieldEditor.registerDefault(StringEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final i = new Field(mirror, const Symbol('i'), int);
  StringEditor editor = new FieldEditor.forField(i);
  expect(editor.isValueReadOnly, isFalse);
  editor.getValue().then(expectAsync1((v) {
    expect(v, equals('7'));
  }));
}

void testStringFieldIntEditorGetValueThrows() {
  FieldEditor.registerDefault(IntEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final s = new Field(mirror, const Symbol('s'), String);
  IntEditor editor = new FieldEditor.forField(s);
  // This should throw because the value 'snarf' cannot be parsed as an int.
  editor.getValue().catchError(expectAsync1((e) {
    expect(e, new isInstanceOf<TransformError>());
  }));
}

void testRectagleFieldStringEditorValueReadOnly() {
  FieldEditor.registerDefault(StringEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final r = new Field(mirror, const Symbol('r'), Rectangle);
  StringEditor editor = new FieldEditor.forField(r);
  // There is no registered String -> Rectangle transform so we expect readonly.
  expect(editor.isValueReadOnly, isTrue);
  editor.getValue().then(expectAsync1((v) {
    expect(v, equals('rectangle(1,2,3,4)'));
  }));
}

void testIntFieldIntEditorSetValue() {
  FieldEditor.register(int, IntEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final i = new Field(mirror, const Symbol('i'), int);
  IntEditor editor = new FieldEditor.forField(i);
  expect(editor.isValueReadOnly, isFalse);
  final verify = expectAsync1((v) {
    expect(v, equals(42));
    expect(obj.i, equals(42));
  });
  editor.setValue(42).then((_) => editor.getValue().then(verify));
}

void testIntFieldStringEditorSetValue() {
  FieldEditor.registerDefault(StringEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final i = new Field(mirror, const Symbol('i'), int);
  StringEditor editor = new FieldEditor.forField(i);
  expect(editor.isValueReadOnly, isFalse);
  final verify = expectAsync1((v) {
    expect(v, equals('42'));
    expect(obj.i, equals(42));
  });
  editor.setValue('42').then((_) => editor.getValue().then(verify));
}

void testIntFieldStringEditorSetValueThrows() {
  FieldEditor.registerDefault(StringEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final i = new Field(mirror, const Symbol('i'), int);
  StringEditor editor = new FieldEditor.forField(i);
  expect(editor.isValueReadOnly, isFalse);
  // This should throw because the value cannot be parsed as an int.
  editor.setValue('42bottlesofbeeronthewall').catchError(expectAsync1((e) {
    expect(e, new isInstanceOf<TransformError>());
  }));
}

void testRectangleFieldStringEditorSetValueThrows() {
  FieldEditor.registerDefault(StringEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final r = new Field(mirror, const Symbol('r'), Rectangle);
  StringEditor editor = new FieldEditor.forField(r);
  // There is no registered String -> Rectangle transform so we expect readonly.
  expect(editor.isValueReadOnly, isTrue);
  // Setting a value when read-only is expected to throw.
  editor.setValue(new Rectangle(5,6,7,8)).catchError(expectAsync1((e) {
    expect(e, new isInstanceOf<UnsupportedError>());
  }));
}

// -----------------------------------------------------------------------------
// Resources
// -----------------------------------------------------------------------------

void testRectangleEditorResources() {
  FieldEditor.register(Rectangle, RectangleEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final r = new Field(mirror, const Symbol('r'), Rectangle);
  FieldEditor editor = new FieldEditor.forField(r);
  expect(editor.isValueReadOnly, isFalse);
  expect(editor, const isInstanceOf<RectangleEditor>());
  expect(editor.resources, const isInstanceOf<RectangleResources>());
  expect(editor.resources.style, equals('SOME CSS'));
}
