
library fields.test.field_list;

import 'package:unittest/unittest.dart';
import 'package:fields/field.dart';
import 'package:fields/field_editor.dart';
import 'package:fields/field_list.dart';

fieldListTest() {
  group('[field_list.basic]', () {
    test('Test the construction of a field list.', 
        testFieldListConstructor);
    test('Test refreshing the source of a field list with default comparator'
        ' and predicate.', testRefreshSourceDefaults);
    test('Test that setting a null predicate on a field list throws.', 
         testSetNullPredicateThrows);
    test('Test that setting a null comparator on a field list throws.', 
        testSetNullComparatorThrows);
    test('Test setting a reverse name comparator on a field list.', 
        testSetReverseNameComparator);
    test('Test setting a predicate that accepts only read-only fields.', 
        testSetReadOnlyPredicate);
    test('Test the combination of setting a reverse name comparator and a'
        ' read-only predicate on a field list.', 
        testSetReverseNameComparatorAndReadOnlyPredicate);
    test('Test setting the value of a field in the current view.', 
        testSetFieldValueInView);
    test('Test setting the value of a field not in the current view.', 
        testSetFieldValueNotInView);
    test('Test refreshing the source of a field list on a dirty object.',
         testRefreshSourceDirtyObject);
  });
  group('[field_list.metadata]', () {
    test('Test that successful parse of object field metadata.', 
         testParseMetadata);
    test('Test that parsing multiple @UseFieldEditor annotations throws.', 
         testParseMultipleUseFieldEditorAnnotationsThrows);
    test('Test that parsing multiple @Name annotations throws.', 
         testParseMultipleNameAnnotationsThrows);    
  });
}

// -----------------------------------------------------------------------------
// Setup
// -----------------------------------------------------------------------------

class HasFields {
  int i = 7;
  final int ri = 42;
  num n = 3.14;
  final num rn = 6.28;
  double d = 1024.4201;
  final double rd = 1.024;
  String s = 'snarf';
  final rs = 'immutable';
  List<int> li = [ 1, 2, 3, 4];
}

class HasMetadata {
  @UseFieldEditor(CustomIntEditor)    
  int customInt = 42;
  @ReadOnly
  double readOnlyDouble = 3.14;  
  @Ignore
  String secret = 'shhhh';
  @Name('goodName')
  String badName = 'rename me please!';
}

class TooManyFieldEditors {
  @UseFieldEditor(CustomIntEditor)
  @UseFieldEditor(CustomIntEditor)
  int customInt = 42;
}

class TooManyNames {
  @Name('one')
  @Name('too many')
  int customInt = 42;
}

class CustomIntEditor extends FieldEditor {
  final Type valueType = int;
  final element = null;
  CustomIntEditor(Field field) : super(field);
}

// -----------------------------------------------------------------------------
// Basic
// -----------------------------------------------------------------------------

void testFieldListConstructor() {
  final list = new FieldList();
  expect(list.comparator, equals(FieldList.DEFAULT_COMPARATOR));
  expect(list.predicate, equals(FieldList.DEFAULT_PREDICATE));
  expect(list, isEmpty);
}

void testRefreshSourceDefaults() {
  final obj = new HasFields();
  final list = new FieldList(); 
  list.refreshSource(obj);  
  expect(list.length, equals(9));  
  expect(list[0].displayName, equals('d'));
  expect(list[1].displayName, equals('i'));
  expect(list[2].displayName, equals('li'));
  expect(list[3].displayName, equals('n'));
  expect(list[4].displayName, equals('rd'));
  expect(list[5].displayName, equals('ri'));
  expect(list[6].displayName, equals('rn'));
  expect(list[7].displayName, equals('rs'));
  expect(list[8].displayName, equals('s'));
}

void testSetNullPredicateThrows() {
  final list = new FieldList(); 
  expect(() {
    list.predicate = null;
  }, throwsA(const isInstanceOf<AssertionError>()));
}

void testSetNullComparatorThrows() {
  final list = new FieldList(); 
  expect(() {
    list.comparator = null;
  }, throwsA(const isInstanceOf<AssertionError>()));
}

void testSetReverseNameComparator() {
  final obj = new HasFields();
  final list = new FieldList();
  list.refreshSource(obj);
  final Comparator reverse = 
      (Field a, Field b) => a.displayName.compareTo(b.displayName)*-1;    
  list.comparator = reverse;
  expect(list.comparator, equals(reverse));
  expect(list.length, equals(9));
  expect(list[0].displayName, equals('s'));
  expect(list[1].displayName, equals('rs'));
  expect(list[2].displayName, equals('rn'));
  expect(list[3].displayName, equals('ri'));
  expect(list[4].displayName, equals('rd'));
  expect(list[5].displayName, equals('n'));
  expect(list[6].displayName, equals('li'));
  expect(list[7].displayName, equals('i'));  
  expect(list[8].displayName, equals('d'));      
}

void testSetReadOnlyPredicate() {
  final obj = new HasFields();
  final list = new FieldList();
  list.refreshSource(obj);
  final Predicate readonly = (Field f) => f.isReadOnly;
  list.predicate = readonly;
  expect(list.predicate, equals(readonly));
  expect(list.length, equals(4));
  expect(list[0].displayName, equals('rd'));
  expect(list[1].displayName, equals('ri'));
  expect(list[2].displayName, equals('rn'));
  expect(list[3].displayName, equals('rs'));
}

void testSetReverseNameComparatorAndReadOnlyPredicate() {
  final obj = new HasFields();
  final list = new FieldList();
  list.refreshSource(obj);
  final Comparator reverse = 
      (Field a, Field b) => a.displayName.compareTo(b.displayName)*-1;
  final Predicate readonly = (Field f) => f.isReadOnly;
  list.comparator = reverse;
  list.predicate = readonly;
  expect(list.length, equals(4));
  expect(list[0].displayName, equals('rs'));
  expect(list[1].displayName, equals('rn'));
  expect(list[2].displayName, equals('ri'));
  expect(list[3].displayName, equals('rd'));
}

void testSetFieldValueInView() {
  final obj = new HasFields();
  final list = new FieldList();
  list.refreshSource(obj);  
  Field i = list[1];
  list.fieldChanged.listen(expectAsync1((f) {
    expect(f, equals(i));
  }));
  i.setValue(42).then(expectAsync1((v) {
    expect(v, equals(42));
  }));
}

void testSetFieldValueNotInView() {
  final obj = new HasFields();
  final list = new FieldList();
  list.refreshSource(obj);
  Field i = list[1];
  final Predicate readonly = (Field f) => f.isReadOnly;
  list.predicate = readonly;
  expect(list, isNot(contains(i)));
  list.fieldChanged.listen((_) => fail('No field in view has changed.'));
  i.setValue(42).then(expectAsync1((v) {
    expect(v, equals(42));
  }));
}

void testRefreshSourceDirtyObject() {
  final obj = new HasFields();
  final list = new FieldList(); 
  list.refreshSource(obj);
  expect(list.length, equals(9));  
  list[0].values.listen(expectAsync1((value) {
    expect(value, equals(3.14));
  }));  
  // Modify a field of the object.
  obj.d = 3.14;  
  // Call to `refreshSource` on dirty object should trigger field value events.
  list.refreshSource(obj);
}

// -----------------------------------------------------------------------------
// Metadata
// -----------------------------------------------------------------------------

void testParseMetadata() {
  final obj = new HasMetadata();
  final list = new FieldList(); 
  list.refreshSource(obj);  
  expect(list.length, equals(3));  
  expect(list[0].name, equals(const Symbol('customInt')));
  // TODO(rms): Caught UnimplementedError: Mirror.hashCode is not implemented
  //expect(list[0].editor, equals(reflectClass(CustomIntEditor)));
  expect(list[1].name, equals(const Symbol('badName')));
  expect(list[1].displayName, equals('goodName'));  
  expect(list[2].name, equals(const Symbol('readOnlyDouble')));
  expect(list[2].isReadOnly, isTrue);  
}

void testParseMultipleUseFieldEditorAnnotationsThrows() {
  final obj = new TooManyFieldEditors();
  final list = new FieldList(); 
  expect(() {
    list.refreshSource(obj);
  }, throwsA(const isInstanceOf<AssertionError>()));  
}

void testParseMultipleNameAnnotationsThrows() {
  final obj = new TooManyNames();
  final list = new FieldList(); 
  expect(() {
    list.refreshSource(obj);
  }, throwsA(const isInstanceOf<AssertionError>()));  
}
