
library fields.test.transform;

import 'package:unittest/unittest.dart';
import 'package:fields/transform.dart';

transformTest() {
  // TODO: we should clear registration in 'setUp' between each test
  group('[transform.basic]', () {
    test('Test that simple types can be transformed.', 
         testCanTransformSimpleTypes);
    test('Test a valid string to int transform.', testTransformStringToInt);
    test('Test an erroneous string to int transform.', 
        testTransformStringToIntThrows);
    test('Test that a string to int transform where value is! a string throws.', 
         testTransformStringToIntValueIsNotStringThrows);    
  });
  group('[transform.custom]', () {
    test('Test the registration of a custom string to int transform.', 
        testRegisterCustomTransformStringToInt);
    test('Test the registration of a custom rectangle transform.', 
         testRegisterRectangleTransform);
    test('Test a valid rectangle transform.', testTransformRectangle);
  });
}

// -----------------------------------------------------------------------------
// Setup
// -----------------------------------------------------------------------------

class CustomStringToIntTransform extends Transform<String, int> {
  Type get sourceType => String;
  Type get targetType => int;
  const CustomStringToIntTransform();
  call(value) => throw new UnimplementedError();
}

class Rectangle {
  int x, y, w, h;
}

class RectangleF {
  double x, y, w, h;
}

class RectangleToRectangleFTransform extends Transform<Rectangle, RectangleF> {
  Type get sourceType => Rectangle;
  Type get targetType => RectangleF;
  const RectangleToRectangleFTransform();
  call(value) => new RectangleF()..x = value.x.toDouble()
      ..y = value.y.toDouble()
      ..w = value.w.toDouble()
      ..h = value.h.toDouble();
}

// -----------------------------------------------------------------------------
// Basic
// -----------------------------------------------------------------------------

void testCanTransformSimpleTypes() {
  expect(canTransform(int, int), isTrue);
  expect(canTransform(String, int), isTrue);
}

void testTransformStringToInt() {
  final String s = '42';
  final int i = transform(s, String, int);
  expect(i, equals(42));
}

void testTransformStringToIntThrows() {
  final String s = '42.123';
  expect(() {
    transform(s, String, int); 
  }, throwsA(const isInstanceOf<TransformError>()));
}

void testTransformStringToIntValueIsNotStringThrows() {
  final map = { 'foo' : 'bar' };
  expect(() { 
    transform(map, String, int); 
  }, throwsA(const isInstanceOf<TypeError>()));
}

// -----------------------------------------------------------------------------
// Custom
// -----------------------------------------------------------------------------

void testRegisterCustomTransformStringToInt() {
  expect(register(const CustomStringToIntTransform()), isFalse);
  expect(canTransform(String, int), isTrue);
}

void testRegisterRectangleTransform() {
  expect(canTransform(Rectangle, RectangleF), isFalse);
  expect(register(const RectangleToRectangleFTransform()), isTrue);
  expect(canTransform(Rectangle, RectangleF), isTrue);
  expect(unregister(const RectangleToRectangleFTransform()), isTrue);
  expect(canTransform(Rectangle, RectangleF), isFalse);
}

void testTransformRectangle() {  
  expect(register(const RectangleToRectangleFTransform()), isTrue);
  final r = new Rectangle()..x = 1..y = 2..w = 3..h = 4;
  final rf = transform(r, Rectangle, RectangleF);
  expect(rf, const isInstanceOf<RectangleF>());
  expect(rf.x, equals(1));
  expect(rf.y, equals(2));
  expect(rf.w, equals(3));
  expect(rf.h, equals(4));
  expect(unregister(const RectangleToRectangleFTransform()), isTrue);
}
